/*

------------------------------------------------------------------------------

A license is hereby granted to reproduce this software source code and
to create executable versions from this source code for personal,
non-commercial use.  The copyright notice included with the software
must be maintained in all copies produced.

THIS PROGRAM IS PROVIDED "AS IS". THE AUTHOR PROVIDES NO WARRANTIES
WHATSOEVER, EXPRESSED OR IMPLIED, INCLUDING WARRANTIES OF
MERCHANTABILITY, TITLE, OR FITNESS FOR ANY PARTICULAR PURPOSE.  THE
AUTHOR DOES NOT WARRANT THAT USE OF THIS PROGRAM DOES NOT INFRINGE THE
INTELLECTUAL PROPERTY RIGHTS OF ANY THIRD PARTY IN ANY COUNTRY.

Copyright (c) 1994-2018, John Conover, All Rights Reserved.

Comments and/or bug reports should be addressed to:

    john@email.johncon.com (John Conover)

------------------------------------------------------------------------------

Smooth the measured frequency response of a speaker. The input file is
sequential text records, in lexical order of frequency, each record
having the frequency, followed by the amplitude at that frequency,
separated by white space. Blank records are ignored, and comment
records are signified by a '#' character as the first non white space
character in the record. The smoothed output is on stdout.

For any record in the file, containing the frequency F, average the
amplitudes between F / 2^(N / 2) to F * 2^(N / 2), where N is 1/2,
1/3, 1/12, etc., fraction of an octave.

Note that Fmin = F / 2^(N / 2), or, F = Fmin * 2^(N / 2) permitting
sequential scanning of the file, i.e., for any frequency record:

    Input the record, the frequency in this record is Fmin; calculate
    F = Fmin * 2^(N / 2).

    Average the amplitudes, Av, in the following records to Fmax = F *
    2^(N / 2).

    print F and Av.

Or, simplifying:

    Input the record, the frequency in this record is Fmin; calculate
    F = Fmin * 2^(N / 2).

    Average the amplitudes, Av, in the following records to Fmax =
    Fmin * 2^N

    print F and Av.

Note that 2^(N / 2) is pow (2, N / 2) from math.h, and a
ftell(3)/fseek(3) construct to the next record will be required.

$Revision: 0.0 $
$Date: 2018/04/12 21:44:42 $
$Id: osmooth.c,v 0.0 2018/04/12 21:44:42 john Exp $
$Log: osmooth.c,v $
Revision 0.0  2018/04/12 21:44:42  john
Initial revision


*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <unistd.h>

#ifdef __STDC__

#include <float.h>

#endif

static char rcsid[] = "$Id: osmooth.c,v 0.0 2018/04/12 21:44:42 john Exp $"; /* program version */
static char copyright[] = "Copyright (c) 1994-2018, John Conover, All Rights Reserved"; /* the copyright banner */

#define BUFLEN BUFSIZ /* i/o buffer size */

#define TOKEN_SEPARATORS " \t\n\r\b," /* file record field separators */

#ifdef __STDC__

static const char *help_message[] = /* help message index array */

#else

static char *help_message[] = /* help message index array */

#endif

{
    "\n",
    "Smooth the measured frequency response of a speaker in a file\n",
    "Usage: osmooth [-n] [-v] [filename]\n",
    "    -n N, N = fraction of an octave to smooth, 0 < N < 1, (0.333... = 1/3)\n",
    "    -v, print the program's version information\n",
    "    filename, input filename\n"
};

#ifdef __STDC__

static const char *error_message[] = /* error message index array */

#else

static char *error_message[] = /* error message index array */

#endif

{
    "No error\n",
    "Error in program argument(s)\n",
    "Error opening file\n",
    "Error closing file\n",
    "Error seeking position in file\n"
};

#define NOERROR 0 /* error values, one for each index in the error message array */
#define EARGS 1
#define EOPEN 2
#define ECLOSE 3
#define ESEEK 4

#ifdef __STDC__

static void print_message (int retval); /* print any error messages */
static int strtoken (char *string, char *parse_array, char **parse, char *delim);

#else

static void print_message (); /* print any error messages */
static int strtoken ();

#endif

#ifdef __STDC__

int main (int argc, char *argv[])

#else

int main (argc, argv)
int argc;
char *argv[];

#endif

{
    char buffer[BUFLEN], /* i/o buffer */
         parsebuffer[BUFLEN], /* parsed i/o buffer */
         *token[BUFLEN / 2], /* reference to tokens in parsed i/o buffer */
         token_separators[] = TOKEN_SEPARATORS;

    int count = 0, /* input file record counter for the fraction of an octive */
        retval = NOERROR, /* return value, assume no error */
        fields, /* number of fields in a record */
        c; /* command line switch */

    long record_position; /* the location in the input file of the record postion of the beginning of the NEXT fractional ocatve */

    double sum = (double) 0.0, /* running value of cumulative sum of a fraction of an octave */
           N = 0.33333333333333333333, /* fraction of an octave to smooth, 1/3 is most like human hearing */
           Fmin = (double) 0.0, /* current lowest frequency to start average, Fmin */
           Fmax = (double) 0.0, /* current highest frequency to end average, Fmax */
           F = (double) 0.0; /* the center frequency, F, between Fmin and Fmax */

    FILE *infile = stdin; /* reference to input file */

    while ((c = getopt (argc, argv, "n:v")) != EOF) /* for each command line switch */
    {

        switch (c) /* which switch? */
        {

            case 'n': /* request to set the fraction of an octave to smooth? */

                N = atof (optarg); /* yes, set the fraction of an octave to smooth */

                if (N <= (double) 0.0 || N >= (double) 1.0) /* fraction of an octave to smooth make sense? */
                {
                    optind = argc; /* force argument error */
                    retval = EARGS; /* assume not enough arguments */
                }

                break;

            case 'v':

                (void) printf ("%s\n", rcsid); /* print the version */
                (void) printf ("%s\n", copyright); /* print the copyright */
                optind = argc; /* force argument error */
                retval = EARGS; /* assume not enough arguments */

            default: /* illegal switch? */

                optind = argc; /* force argument error */
                retval = EARGS; /* assume not enough arguments */
                break;
        }

    }

    if (retval == NOERROR) /* enough arguments? */
    {
        retval = EOPEN; /* assume error opening file */

        if ((infile = argc <= optind ? stdin : fopen (argv[optind], "r")) != (FILE *) 0) /* yes, open the input file */
        {
            retval = NOERROR; /* assume no error */
            count = 0; /* reset the input file record counter for this fraction of an octive */
            sum = (double) 0.0; /* reset running value of cumulative sum of a fraction of an octave */

            while (fgets (buffer, BUFLEN, infile) != (char *) 0) /* read the records from the input file */
            {

                if ((fields = strtoken (buffer, parsebuffer, token, token_separators)) != 0) /* parse the record into fields, skip the record if there are no fields */
                {

                    if (token[0][0] != '#') /* if the first character of the first field is a '#' character, skip it */
                    {

                        if (count == 0) /* first record of a fraction of octave sequence? */
                        {
                            Fmin = atof (token[0]); /* yes, save the current lowest frequency to start average, Fmin */
                            F = Fmin * pow (2, N / 2); /* save the center frequency, F, between Fmin and Fmax */
                            Fmax = Fmin * pow (2, N); /* save current highest frequency to end average, Fmax */

                            /*

                            first record of a fractional octave to be averaged
                            has been read from the input file; the file
                            pointer of the input file is pointed at the second
                            record the fractional octave to be averaged, which
                            will be the first record of the NEXT octave to be
                            averaged

                            */

                            record_position = ftell (infile); /* save the location in the input file of the record postion of the beginning of the NEXT fractional ocatve */

                            /* (void) printf ("#1 Fmin = %f\tF = %f\tFmax = %f\tsum = %f\tcount = %d\tnext position = %li\n", Fmin, F, Fmax, sum, count, record_position); */
                        }

                        if (atof (token[0]) <= Fmax) /* frequency represented by this record <= Fmax, (i.e., NOT end of fractional octave)? */
                        {
                            sum = sum + atof (token[fields - 1]); /* save the value of the current sample in the time series, and add it to the running value of cumulative sum for this fraction of an octave */
                            count ++; /* increment the count of records from the input file for this fractional octave */
                        }

                        else
                        {

                            /*

                            end of this fractional octave:

                            */

                            /* (void) printf ("#2 sum = %f\tcount = %i\n", sum, count); */

                            (void) printf ("%f\t%f\n", F, sum / (double) count); /* print the average of this fraction of an octave */

                            count = 0; /* reset the input file record counter for this fraction of an octive */
                            sum = (double) 0.0; /* reset running value of cumulative sum of a fraction of an octave */

                            retval = ESEEK; /* assume error seeking position in file */

                            if (fseek (infile, record_position, SEEK_SET) == 0)
                            {
                                retval = NOERROR; /* assume no error */
                            }

                        }

                    }

                }

            }

            if (argc > optind) /* using stdin as input? */
            {

                if (fclose (infile) == EOF) /* no, close the input file */
                {
                    retval = ECLOSE; /* error closing file */
                }

            }

        }

    }

    print_message (retval); /* print any error messages */
    exit (retval); /* exit with the error value */

#ifdef LINT

    return (0); /* for lint formalities */

#endif

}

/*

Print any error messages.

static void print_message (int retval);

I) Data structures:

    A) The help_message array is an array of character stings, one per
       line to be printed for help.

    B) The error_message array is an array of character strings, one
       line per error; the error_message array is implicitly addressed
       by the value integer retval, which specifies the error to be
       printed.

II) Function execution:

    A) Depending on the value of the integer retval:

        1) If retval is zero, print nothing-normal/successful program
           exit.

        2) Else if retval is unity, print help.

        3) Else retval is not zero or unity, it is an error code, print
           the corresponding error message.

Returns nothing.

*/

#ifdef __STDC__

static void print_message (int retval)

#else

static void print_message (retval)
int retval;

#endif

{
    size_t help_ctr; /* help_message line counter */

    switch (retval) /* which return value */
    {

        case 0: /* program ended without errors, print nothing */

            break;

        case 1: /* program ended with a request for help, or argument error, print help */

            for (help_ctr = 0; help_ctr < (sizeof (help_message) / sizeof (char *)); help_ctr ++) /* for each line of help */
            {
                (void) printf ("%s", help_message[help_ctr]); /* print the line */
            }

            break;

        default: /* an error that was not a request for help, print the error */

            (void) fprintf (stderr, "%s", error_message[retval]);
            break;

    }

}

/*

parse a record based on sequential delimiters

int strtoken (char *string, char *parse_array, char **parse, char *delim);

parse a character array, string, into an array, parse_array, using
consecutive characters from delim as field delimiters, point the
character pointers, token, to the beginning of each field, return the
number of fields parsed

*/

#ifdef __STDC__

static int strtoken (char *string, char *parse_array, char **parse, char *delim)

#else

static int strtoken (string, parse_array, parse, delim)
char *string,
*parse_array,
**parse,
*delim;

#endif

{
    int tokens = 0;

    (void) strcpy (parse_array, string); /* copy the string */

    parse[tokens] = strtok (parse_array, delim); /* get the 1st field */

    while (parse[tokens] != 0) /* get the remaining fields */
    {
        parse[++ tokens] = strtok ((char *) 0, delim);
    }

    return (tokens); /* return the number of tokens parsed */
}
